from ..base import *
import json
import httpx
import requests
import urllib.parse
from time import sleep
# from typing import Union

def get_catalog(board: Board, as_dict: bool = False) -> List[CatalogThread]:
    url = f'https://a.4cdn.org/{Board[board]}/catalog.json'
    data = requests.get(url).json()

    all_posts: List[CatalogThread] = []
    for page in data:
        for thread in page['threads']:
            '''attach board to thread'''
            assert not isinstance(board, List), 'board should not be list'
            thread['board'] = board
            # if as_dict:
            # all_posts.append(CatalogThread(**thread).dict())
            # else:
            all_posts.append(CatalogThread(**thread))
    return all_posts



def catalog_image_generator(board: Board):
    url = f'https://a.4cdn.org/{board}/catalog.json'
    r = requests.get(url).json()
    lst = []
    for idx, page in enumerate(r):
        for thread in r[idx]['threads']:
            if 'last_replies' in thread:
                for comment in thread['last_replies']:
                    if 'ext' in comment and 'tim' in comment:
                        url = 'http://i.4cdn.org/{0}/{1}{2}'.format(
                            board, 
                            str(comment['tim']), 
                            str(comment['ext'])
                        )
                        lst.append(url)
    print(lst)
    print(len(lst))
    for i in lst:
        yield i

def iter_img_lst():
    counter = 0
    for img in catalog_image_generator(Board.pol):
        sleep(1.01)
        file_name = img.split('/')[-1]
        filename, headers = urllib.request.urlretrieve(img, f'./chan_images/{file_name}')
        counter += 1
        print(f'{counter}: {filename} {headers}')





























class AsyncHTTP:

    def get_or_create_eventloop(self):
        # https://techoverflow.net/2020/10/01/how-to-fix-python-asyncio-runtimeerror-there-is-no-current-event-loop-in-thread/
        try:
            return asyncio.get_event_loop()
        except RuntimeError as err:
            if "There is no current event loop in thread" in str(err):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                return asyncio.get_event_loop()

    async def _read_lines(self, url: str, client: httpx.AsyncClient):
        assert len(url) > 10
        try:
            async with client.stream("GET", url) as resp:
                async for line in resp.aiter_lines():
                    yield json.loads(line)
        except httpx.RemoteProtocolError as e:
            print('read_lines: ', e)
            yield e


async def get_catalog_v2():
    url = 'https://a.4cdn.org/wg/catalog.json'
    client = httpx.AsyncClient()
    aa = AsyncHTTP()
    async_gen = aa._read_lines(url, client)

    all_posts = []
    # _type = ''
    async for page in async_gen:
        # _type = type(page)
        for thread in page:
            for item in thread['threads']:
                all_posts.append(item)

            # all_posts.append(thread)
            # thread = jsonable_encoder(thread)
            # all_posts.append(CatalogThread(**thread))
    # print(_type)
    await async_gen.aclose()
    await client.aclose()
    return all_posts



from .base import *
import strawberry

@strawberry.enum
class Board(str, Enum):
    wg = 'wg'
    v = 'v'
    b = 'b'
    pol = 'pol'
    s = 's'
    h = 'h'
    c = 'c'
    e = 'e'
    g = 'g'
    k = 'k'
    o = 'o'
    u = 'u'
    vg = 'vg'
    r9k = 'r9k'
    s4s = 's4s'
    cm = 'cm'
    hm = 'hm'
    lgbt = 'lgbt'
    sci = 'sci'
    wsg = 'wsg'
    adv = 'adv'
    an = 'an'
    out = 'out'
    trv = 'trv'
    sp = 'sp'
    soc = 'soc'
    fit = 'fit'
    biz = 'biz'
    fa = 'fa'
    tg = 'tg'
    w = 'w'
    x = 'x'
    err = 'err'

class FourChanBase(BaseModel):
    pass   
    # @classmethod
    # async def list_threads(cls, db: AsyncIOMotorDatabase):
    #     # project = {'_id': 0, 'no': 1, 'com': 1, 'last_replies': 1}
    #     project: dict = {}
    #     q = db.find({}, project, limit=10)
    #     return [cls(**i) async for i in q]
    #     # return [CatalogBase(**i) async for i in q]

class CatalogBase(FourChanBase):
    no: Optional[int] = Field(None, description="always | The numeric post ID | any positive integer")
    resto: Optional[int] = Field(None, description="always | For replies: this is the ID of the thread being replied to. For OP: this value is zero |`0` or `Any positive integer")
    sticky: Optional[int] = Field(None, description="OP only, if thread is currently stickied | If the thread is being pinned to the top of the page| `1` or not set")
    closed: Optional[int] = Field(None, description="OP only, if thread is currently closed | If the thread is closed to replies | `1` or not set")
    now: Optional[str] = Field(None, description="always | MM/DD/YY(Day)HH:MM (:SS on some boards), EST/EDT timezone | `string")
    time: Optional[int] = Field(None, description="always | UNIX timestamp the post was created | UNIX timestamp")
    name: Optional[str] = Field(None, description="always | Name user posted with. Defaults to Anonymous | any string")
    trip: Optional[str] = Field(None, description="if post has tripcode | The users tripcode, in format: !tripcode or !!securetripcode| any string")
    id: Optional[str] = Field(None, description="if post has ID | posters ID | any 8 chars")

    capcode: Optional[str] = Field(None, description="if post has capcode | The capcode identifier for a post | Not set, mod, admin, admin_highlight, manager, developer, founder")

    country: Optional[str] = Field(None, description="if country flags are enabled | Posters [ISO 3166-1 alpha-2 country code](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2 | 2 character string or XX if unknown")
    country_name: Optional[str] = Field(None, description="if country flags are enabled | Posters country name | Name of any country")

    sub: Optional[str] = Field(None, description="OP only, if subject was included| OP Subject text | any string")

    com: Optional[str] = Field(None, description="if comment was included | Comment (HTML escaped) | any HTML escaped string")

    tim: Optional[int] = Field(None, description="always if post has attachment | Unix timestamp + microtime that an image was uploaded | integer")
    filename: Optional[str] = Field(None, description="always if post has attachment | Filename as it appeared on the poster's device | any string")
    ext: Optional[str] = Field(None, description="always if post has attachment | Filetype | jpg, png, gif, pdf, swf, webm")
    fsize: Optional[int] = Field(None, description="always if post has attachment | Size of uploaded file in bytes | any integer")
    md5: Optional[str] = Field(None, description="always if post has attachment | 24 character, packed base64 MD5 hash of file")
    w: Optional[int] = Field(None, description="always if post has attachment | Image width dimension | `any integer")
    h: Optional[int] = Field(None, description="always if post has attachment | Image height dimension | `any integer")
    tn_w: Optional[int] = Field(None, description="always if post has attachment | Thumbnail image width dimension | any integer")
    tn_h: Optional[int] = Field(None, description="always if post has attachment | Thumbnail image height dimension | any integer")
    filedeleted: Optional[int] = Field(None, description="if post had attachment and attachment is deleted | If the file was deleted from the post | `1` or not set")
    spoiler: Optional[int] = Field(None, description="if post has attachment and attachment is spoilered | If the image was spoilered or not | `1` or not set")
    custom_spoiler: Optional[int] = Field(None, description="if post has attachment and attachment is spoilered | The custom spoiler ID for a spoilered image | `1-10` or not set |")
    omitted_posts: Optional[int] = Field(None, description="OP only| Number of replies minus the number of previewed replies | `any integer` |")
    omitted_images: Optional[int] = Field(None, description="OP only| Number of image replies minus the number of previewed image replies | `any integer` |")
    replies: Optional[int] = Field(None, description="OP only | Total number of replies to a thread | any integer")
    images: Optional[int] = Field(None, description="OP only | Total number of image replies to a thread | any integer")
    bumplimit: Optional[int] = Field(None, description="OP only, only if bump limit has been reached | If a thread has reached bumplimit, it will no longer bump | `1` or not set |")
    imagelimit: Optional[int] = Field(None, description="OP only, only if image limit has been reached | If an image has reached image limit, no more image replies can be made  | `1` or not set |")
    last_modified: Optional[int] = Field(None, description="OP only | UNIX timestamp marking last time thread was modified post | added/modified/deleted, thread closed/sticky settings modified | `UNIX Timestamp")
    tag: Optional[str] = Field(None, description="OP only, /f/ only | The category of `.swf` upload |`Game`, `Loop`, etc")
    semantic_url: Optional[str] = Field(None, description="OP only | SEO URL slug for thread | `string` |")
    since4pass: Optional[int] = Field(None, description="if poster put 'since4pass' in the options field` | Year 4chan pass bought | `any 4 digit year`")
    unique_ips: Optional[int] = Field(None, description="OP only | Number of unique posters in a thread | any integer")
    m_img: Optional[int] = Field(None, description="any post that has a mobile-optimized image` | Mobile optimized image exists for post | 1 or not set")

class CatalogThread(CatalogBase):
    board: Board
    last_replies: List[CatalogBase] = Field([], description='Thread replies.')# catalog OP only | JSON representation of the most recent replies to a thread | array of JSON post objects")

@strawberry.experimental.pydantic.type(
    model=CatalogBase, 
    all_fields=True)
class CatalogBaseGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=CatalogThread, 
    all_fields=True)
class CatalogThreadGQL:
    pass


from functools import lru_cache
from pydantic import BaseSettings
from os import environ
from typing import List

class _BaseSettings(BaseSettings):
    # INVALID_DOC_ERR: Callable = lambda x: f"{x}: Text PDF or Invalid File"
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types
    ACCEPT_PDF: List[str] = [".pdf"]
    ACCEPT_IMAGE: List[str] = ['.jpg', '.png']
    ACCEPT_FILES: List[str] = ACCEPT_IMAGE + ACCEPT_PDF
    FILE_CONVERSION_LIST: List[str] = ['.jpeg']

    WHICH_LOGGER: str= environ.get('WHICH_LOGGER')
    
    APP_NAME: str = environ.get("APP_NAME")
    APP_VERSION: str = environ.get("APP_VERSION")
    ENV_NAME: str = environ.get("ENV_NAME")

    GMAIL_APP_PASSWORD: str = environ.get("GMAIL_APP_PASSWORD")
    TEST_EMAIL: str = environ.get("TEST_EMAIL")


    INPUT_BUCKET: str = environ.get('INPUT_BUCKET')
    AWS_REGION: str = environ.get('AWS_REGION')
    MONGO_URI: str = environ.get('MONGO_URI')
    MONGO_DB_NAME: str = environ.get('MONGO_DB_NAME')
    # ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
    # DATA_DIR = os.path.join(ROOT_DIR, 'data')
    # COMPANY_ENUM_FILE: str = environ.get('COMPANY_ENUM_FILE')
    # BASE_DIR_FILES: Path = environ.get('BASE_DIR_FILES')

@lru_cache()
def get_settings() -> BaseSettings:
    return _BaseSettings()



from .base import *

datetime_to_date = lambda t: t.isoformat(' ', 'hours').split(' ')[0]
datetime_to_str = lambda t: t.isoformat()
datetime_to_unix_epoch = lambda x: x.timestamp()

class BaseConfig(BaseModel):

    def to_dict(self, **kwargs):
        data = self.dict(by_alias=True)
        data["_id"] = str(self.id)
        return data

    class Config:
        allow_population_by_field_name = True
        arbitrary_types_allowed = True
        json_encoders = {
            datetime: datetime_to_unix_epoch,
            bson.decimal128.Decimal128: lambda x: str(x),
            bson.ObjectId: str
        }



from ..base_db import *
from typing import List, Dict, Any, Callable
from pydantic import BaseModel, AnyUrl
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase
from pymongo import ( 
    InsertOne,
    ReplaceOne
)

class MongoCrud(object):

    def __init__(self, uri: AnyUrl, db_name: str) -> None:
        """Motor provides a single client class, MotorClient. 
        Unlike PyMongos MongoClient, Motors client class does not begin connecting in the background when it is instantiated. 
        Instead it connects on demand, when you first attempt an operation.

        https://motor.readthedocs.io/en/stable/differences.html#connecting-to-mongodb
        Args:
            uri (AnyUrl): _description_
            db_name (str): _description_
        """
        client = AsyncIOMotorClient(uri, uuidRepresentation="standard")
        self.db = client[db_name]

    def get_mongo_db(self) -> AsyncIOMotorDatabase:
        """async Motor db connection

        Returns:
            AsyncIOMotorDatabase: _description_

        Yields:
            Iterator[AsyncIOMotorDatabase]: _description_
        """
        '''this throws TypeError: 'AsyncIOMotorDatabase' object is not iterable'''
        return self.db

    @property
    def get_db_name(self) -> Any:
        return self.db.name

    @property
    async def get_build_info(self) -> Dict:
        return await self.db.command("buildinfo")

    @property
    async def get_collection_names(self) -> List[str]:
        non_system = {"name": {"$regex": r"^(?!system\.)"}}
        return await self.db.list_collection_names(filter=non_system)

    async def create_unique_idx(self, db, field: str) -> Any:
        """a missing property will be added to the index as if it the value of that property were null.

        Args:
            db (_type_): _description_
            field (str): _description_

        Returns:
            Any:
        """
        res = await db.create_index(field, unique=True)
        return res

    async def create_multi_index(self, db) -> Any:
        pass
        # from pymongo import IndexModel, ASCENDING, DESCENDING
        # index1 = IndexModel(
        #     [
        #         ("hello", DESCENDING), 
        #         ("world", ASCENDING)
        #     ], 
        #     name="hello_world"
        # )
        # index2 = IndexModel([("goodbye", DESCENDING)])
        # resp = await db.create_indexes([index1, index2])
        # return ''

    # async def create_collection(self, db, name):
        # return await db.create_collection(name)

    # async def count_docs(self, db) -> int:
        # count = await db.count_documents({})
        # return count


    async def drop_collection(self, db, collection_name) -> Any:
        dropped = await db.drop_collection(collection_name)
        return dropped

    async def get_distinct(self, db: AsyncIOMotorDatabase, field: str) -> Any:
        """
        - Distinct values for a given key

        https://www.mongodb.com/docs/manual/core/transactions/#distinct-operation

        Args:
            db (AsyncIOMotorDatabase): _description_
            field (str): _description_

        Raises:
            Exception: Any

        Returns:
            Any: List of unique values
        """
        try:
            cur = await db.distinct(field)
            return cur
        except Exception as err:
            logger.info(f'{err}')
            raise Exception(f'{err}')

    async def delete_all(
        self, 
        db,
        obj: dict
        ) -> Any:
        try:
            return await db.delete_many(obj)
        except Exception as e:
            return e

    async def insert_many_documents(
        self, 
        db, 
        collection_name, 
        results: List, 
        is_testing: bool = False
        ) -> Any:
        '''
        https://pymongo.readthedocs.io/en/stable/examples/bulk.html#ordered-bulk-write-operations
        '''
        if is_testing:
            try:
                await db.drop_collection(collection_name)
            except Exception as err:
                logger.info(f'insert many: {err}')

        await db[collection_name].insert_many((i for i in results))
        # count = await self.count_docs(db[collection_name])
        # logger.info({'Total Document Count': count})
        return 'insert many success'

    async def _insert_docs_test(
        self,
        db,
        collection_name,
        results: List,
        ) -> Any:
        result = await db[collection_name].insert_many(({'x': i} for i in range(2)))
        r = result.inserted_ids
        print(r)
        return r

    async def find_all(
        self, 
        db, 
        model=None,
        _filter=None, 
        _projection=None, 
        as_dict=False,
        skip=0, 
        limit=0) -> List:
        query = db.find(_filter, _projection, skip=0, limit=0)
        if model:
            if as_dict:
                return [model(**raw).dict() async for raw in query]
            return [model(**raw) async for raw in query]
        return [raw async for raw in query]


    async def insert_or_update_many(
        self, 
        db_from: AsyncIOMotorDatabase, 
        db_to: AsyncIOMotorDatabase, 
        model: Callable, 
        _filter: dict = {}
        ) -> Any:
        assert issubclass(model, BaseModel)
        try:
            query = db_from.find(filter=_filter)
            requests = [
                ReplaceOne(
                    {'_id': raw['_id']},
                    model(**raw).dict(),
                    upsert=True
                ) async for raw in query
            ]
            result = await db_to.bulk_write(requests)
            res_str = f'''New:\n-{result.upserted_ids}\nModified:\n-{result.modified_count}'''
            logger.info(res_str)
            return res_str
        except Exception as err:
            logger.warn(f'Update Failed: {err}')
            return err


    async def bulk_insert_one(
        self, 
        db_from: AsyncIOMotorDatabase, 
        db_to: AsyncIOMotorDatabase, 
        model: Callable, 
        _filter: dict = None
        ) -> Any:
        try:
            query = db_from.find(filter=_filter)
            requests = [
                InsertOne(
                    # model(**raw).dict()
                    model(**raw)
                ) async for raw in query
            ]
            result = await db_to.bulk_write(requests)
            res_str = f'''New:\n-{result.upserted_ids}\nModified:\n-{result.modified_count}'''
            logger.info(res_str)

            return res_str
        except Exception as err:
            logger.warn(f'Update Failed: {err}')
            return err


    async def aggregation(self, db, pipe_line: List, model: Callable = None) -> List:
        try:
            cur = db.aggregate(pipe_line, allowDiskUse=True)
            if model:
                return [model(**i).dict() async for i in cur]
            return [i async for i in cur]
        except Exception as err:
            logger.info(f'{err}')
            raise Exception(f'{err}')


    async def create(
        self, 
        db,
        obj: dict
        ):
        try:
            result = await db.insert_one(obj)
            return result
        except Exception as e:
            return e

    async def update(
        self, 
        db,
        query_by: dict,
        update_by: dict, 
        ):
        # p. 207
        try:
            return await db.update_one(query_by, update_by)
        except Exception as e:
            return e

    async def delete(
        self, 
        obj: dict, 
        db):
        # p. 208
        try:
            return await db.delete_one(obj)
        except Exception as e:
            return e

    # async def read(
    #     self, 
    #     db,
    #     search_by = None,
    #     filter_by = None,
    #     # model = None, 
    #     ):
    #     """
    #     Read Document with .find()

    #     https://pymongo.readthedocs.io/en/stable/api/pymongo/collection.html#pymongo.collection.Collection.find

    #     Args:
    #         db (_type_): _description_
    #         search_by (_type_, optional): _description_. Defaults to None.
    #         filter_by (_type_, optional): _description_. Defaults to None.

    #     Returns:
    #         _type_: _description_
    #     """
    #     try:
    #         doc = db.find(search_by, filter_by)
    #         # if model:
    #             # return [model(**raw) async for raw in doc]
    #         return [raw async for raw in doc]
    #     except Exception as e:
    #         return e






from ..config import *
from enum import Enum, auto
import re
from uuid import uuid4
from pydantic import root_validator
import strawberry

'''
published:  Fri, 04 Mar 2022 23:20:00 +0000
published_parsed:  time.struct_time(tm_year=2022, tm_mon=3, tm_mday=4, tm_hour=23, tm_min=20, tm_sec=0, tm_wday=4, tm_yday=63, tm_isdst=0)
authors:  [{'name': 'Tyler Durden'}]
author:  Tyler Durden
author_detail:  {'name': 'Tyler Durden'}
id:  731160 at https://www.zerohedge.com
guidislink:  False
title:  Coinbase CEO: "Ordinary Russians Are Using Crypto As A Lifeline"
title_detail:  {
    'type': 'text/plain', 
    'language': None, 
    'base': 'https://cms.zerohedge.com/', 
    'value': 'Coinbase CEO: "Ordinary Russians Are Using Crypto As A Lifeline"'}
links:  [{'rel': 'alternate', 'type': 'text/html', 'href': 'https://www.zerohedge.com/crypto/coinbase-ceo-ordinary-russians-are-using-crypto-lifeline'}]
link:  https://www.zerohedge.com/crypto/coinbase-ceo-ordinary-russians-are-using-crypto-lifeline
summary:  <span class="field field--name-title field--type-string field--label-hidden">Coinbase CEO: "Ordinary Russians Are Using Crypto As A Lifeline"</span>

[{'author': 'Tyler Durden',
  'id': '733984 at https://www.zerohedge.com',
  'link': 'https://www.zerohedge.com/political/photos-show-miami-beach-transformed-ghost-town-after-curfews',
  'links': [{'href': 'https://www.zerohedge.com/political/photos-show-miami-beach-transformed-ghost-town-after-curfews',
             'rel': 'alternate',
             'type': 'text/html'}],
  'published': 'Sun, 27 Mar 2022 18:00:00 +0000',
  'title': "Photos Show Miami Beach Transformed Into Ghost Town After 'Spring "
           "Break' Curfews"}]

  'summary_detail': {'base': 'https://cms.zerohedge.com/',
                     'language': None,
                     'type': 'text/html',
                     'value': '<span class="field field--name-title '
                              'field--type-string field--label-hidden">How The '
                              "West's Ban On Russian Gold Could "
                              'Backfire</span>
'''

class EnumBase(Enum):
    @classmethod
    def list_name_or_value(cls, name_or_value: str) -> List[str]:
        """List names or values of derived enum class

        Args:
            name_or_value (str): name  | value

        Returns:
            (List[str]): list of derived enum classes names or values
        """
        return [getattr(i, name_or_value) for i in cls.__members__.values()]

    @classmethod
    def list_values_to_titlecase(cls):
        def to_title_case(s):
            return re.sub(r"[A-Za-z]+('[A-Za-z]+)?", lambda mo: mo.group(0).capitalize(), s)
        return [to_title_case(i.value.replace('_', ' ')) for i in cls.__members__.values()]


class EnumAutoBase(EnumBase):
    def _generate_next_value_(name, start, count, last_values):
        return name

@strawberry.enum
class EnumRSS(str, EnumBase):
    cbc_world = 'cbc.ca/cmlink/rss-world'
    cbc_health = 'https://rss.cbc.ca/lineup/health.xml'
    full_rss = 'https://www.zerohedge.com/fullrss2.xml'
    # zh = 'http://feeds.feedburner.com/zerohedge/feed'
    pf = 'https://www.reddit.com/r/ActualPublicFreakouts.rss'
    cnn = 'http://rss.cnn.com/rss/cnn_topstories.rss'
    # York Times Home Page
    nyt = 'http://feeds.nytimes.com/nyt/rss/HomePage'
    nyt_daily = 'https://feeds.simplecast.com/54nAGcIl'
    
    the_atlantic = 'https://www.theatlantic.com/feed/all'
    
    nbc_news = 'http://www.wthr.com/Global/category.asp?C=79076&clienttype=rss'
    # washhington Post: Today's Highlights
    wp = 'http://www.washingtonpost.com/rss/'
    # Top U.S. News
    ap = 'https://www.apnews.com/apf-usnews'
    # TODAY.com News - Top Stories
    usat = 'http://rssfeeds.usatoday.com/usatoday-NewsTopStories'
    #  Topics: News
    npr = 'http://www.npr.org/rss/rss.php?id=1001'
    #  News - Americas: World Edition
    # ps://www.bbc.co.uk/news/10628494#userss
    bbc = 'http://newsrss.bbc.co.uk/rss/newsonline_world_edition/americas/rss.xml'
    # ps://www.cbc.ca/rss/

    ctv_canada = 'http://ctvnews.ca/rss/Canada'
    ctv_politics = 'http://www.ctvnews.ca/rss/Politics'
    ctv_top = 'http://ctvnews.ca/rss/TopStories'
    ctv_bc = 'http://bc.ctvnews.ca/rss/bcnews'
    ctv_calgary = 'http://calgary.ctvnews.ca/rss/CalgaryNews'
    ctv_toronto = 'http://toronto.ctvnews.ca/rss/Latest'
    ctv_atlantic = 'http://atlantic.ctvnews.ca/rss/Atlantic'

    # https://www.thetelegraph.com/rss/
    telegraph = 'https://www.thetelegraph.com/rss/feed/News-RSS-Feed-1978.php'

    yahoo_news = 'https://news.yahoo.com/rss/'

    # https://www.dailymail.co.uk/home/article-2684527/RSS-Feeds.html
    daily_mail_latest = 'https://www.dailymail.co.uk/articles.rss'
    daily_mail_health = 'https://www.dailymail.co.uk/health/index.rss'
    daily_mail_science = 'https://www.dailymail.co.uk/sciencetech/index.rss'
    daily_mail_az = 'https://www.dailymail.co.uk/news/astrazeneca/index.rss'
    daily_mail_asia = 'https://www.dailymail.co.uk/news/asia/index.rss'
    dm_bill_gates = 'https://www.dailymail.co.uk/news/bill-gates/index.rss'
    dm_canada = 'https://www.dailymail.co.uk/news/canada/index.rss'
    dm_covid = 'https://www.dailymail.co.uk/news/coronavirus/index.rss'
    dm_depression = 'https://www.dailymail.co.uk/news/depression/index.rss'

    dm_fbi = 'https://www.dailymail.co.uk/news/fbi/index.rss'
    dm_giz_lane = 'https://www.dailymail.co.uk/news/ghislainemaxwell/index.rss'

@strawberry.enum
class Topic(str, EnumAutoBase):
    UN_ASSIGNED = auto()

@strawberry.enum
class PropCategory(str, EnumAutoBase):
    COVERUP = auto()
    PROP_AGANDA = auto()
    CO_ORDINATED = auto()
    UN_CATEGORIZED = auto()


class Authors(BaseConfig):
    name: Optional[str] = None

class AuthorDetail(Authors):
    pass

class TitleDetail(BaseConfig):
    base: Optional[str] = None
    language: Optional[str] = None
    type: Optional[str] = None
    value: Optional[str] = None
    class Config:
        schema_extra = {
            'type': 'text/plain', 
            'language': None, 
            'base': 'https://cms.zerohedge.com/', 
            'value': 'Coinbase CEO: "Ordinary Russians Are Using Crypto As A Lifeline"'
        }

class Links(BaseConfig):
    rel: Optional[str] = None
    type: Optional[str] = None
    href: Optional[str] = None

    class Config:
        schema_extra = {
            'links': [{'rel': 'alternate', 'type': 'text/html', 'href': 'https://www.zerohedge.com/crypto/coinbase-ceo-ordinary-russians-are-using-crypto-lifeline'}]
        }

class MediaCredit(BaseConfig):
    content: Optional[str] = None


class MediaContent(BaseConfig):
    medium: Optional[str] = None
    url: Optional[str] = None
    height: Optional[str] = None
    width: Optional[str] = None
    type: Optional[str] = None

class Tags(BaseConfig):
    term: Optional[str] = None
    scheme: Optional[str] = None
    label: Optional[str] = None


class SummaryDetail(TitleDetail):
    '''SummaryDetail has same fields as TitleDetail'''
    pass

class RSS_Schema(BaseConfig):
    id: str = Field(default_factory=uuid4, alias="_id")# type: ignore
    unique_id: Optional[str] = Field('')
    authors: Optional[List[Authors]] = None
    author: Optional[str] = None
    author_detail: Optional[AuthorDetail] = None
    credit: Optional[str] = None
    media_credit: Optional[List[MediaCredit]] = None
    title: Optional[str] = None
    link: Optional[str] = None
    links: Optional[List[Links]] = None
    published: Optional[str] = None
    tags: Optional[List[Tags]] = None
    media_content: Optional[List[MediaContent]] = None
    summary_detail: Optional[SummaryDetail] = None
    flag: PropCategory = PropCategory.UN_CATEGORIZED
    topic: Topic = Topic.UN_ASSIGNED
    is_sorted: bool = False
    rss_url: str

    @root_validator(pre=True)
    def create_uid(cls, val):
        # gql wants alias _id not model name id
        val['unique_id'] = str(val['rss_url']) + '-' + str(val['link'])
        return val

@strawberry.experimental.pydantic.type(
    model=Authors, 
    all_fields=True)
class AuthorsGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=AuthorDetail, 
    all_fields=True)
class AuthorDetailGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=TitleDetail, 
    all_fields=True)
class TitleDetailGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=Tags, 
    all_fields=True)
class TagsGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=MediaCredit, 
    all_fields=True)
class MediaCreditGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=MediaContent, 
    all_fields=True)
class MediaContentGQL:
    pass


@strawberry.experimental.pydantic.type(
    model=Links, 
    all_fields=True)
class LinksGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=SummaryDetail, 
    all_fields=True)
class SummaryDetailGQL:
    pass

@strawberry.experimental.pydantic.type(
    model=RSS_Schema, 
    all_fields=True)
class RSS_SchemaGQL:
    pass





# from .crowdfund.crowd_fund import *
# from .farm_army.farm_army import *
# from .blockchain.eth_api import *
# from .google_trends.trends import *
from .four_chan.four_chan import *
from .rss.rss_feed import RSSFeed





import feedparser
from ..base import *

class RSSFeed(object):
    def get_feed(self, url: str) -> List[RSS_Schema]:
        d = feedparser.parse(url)
        rss = []
        for k in d['entries']:
            post = {}
            for k, v in k.items():
                post[k] = v
                post['rss_url'] = url
            rss.append(RSS_Schema(**post))
        return rss



import logging
from os import environ

def setup_logger(which_logger):
    logging.basicConfig(
        filename=environ.get('LOG_FILE_PATH'),# taken from loca .env file, not set in settings.py
        format='%(asctime)s - %(process)d-%(levelname)s-%(funcName)s - %(message)s', 
        datefmt='%d-%b-%y %H:%M:%S',
        level=logging.INFO)
    log = logging.getLogger(which_logger)
    return log


WHICH_LOGGER = environ.get('WHICH_LOGGER')
# assert WHICH_LOGGER is not None, 'Set WHICH_LOGGER'
logger = setup_logger(WHICH_LOGGER)






# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['osint_tools',
 'osint_tools.api',
 'osint_tools.api.email',
 'osint_tools.api.email.gmail',
 'osint_tools.api.four_chan',
 'osint_tools.api.rss',
 'osint_tools.db',
 'osint_tools.db.mongo_db',
 'osint_tools.schemas',
 'osint_tools.schemas.rss']

package_data = \
{'': ['*']}

install_requires = \
['feedparser>=6.0.8,<7.0.0',
 'motor>=3.0.0,<4.0.0',
 'pydantic>=1.10.1,<2.0.0',
 'python-jose>=3.3.0,<4.0.0',
 'requests>=2.11',
 'strawberry-graphql[debug-server,fastapi]>=0.139.0,<0.140.0']

setup_kwargs = {
    'name': 'osint-tools',
    'version': '0.2.0',
    'description': '',
    'long_description': '\nTools for working with data.\n',
    'author': 'Alexander Slessor',
    'author_email': 'alexjslessor@gmail.com.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




from ...base import *
from imaplib import IMAP4_SSL
import email
from email.header import decode_header
import re
from datetime import datetime
from contextlib import contextmanager
import os
from enum import Enum

'''Instructions:
1. create app password
https://support.google.com/accounts/answer/185833
'''
@contextmanager
def IMAP4_conn_manager(
    host: str = None,
    account_email: str = None, 
    app_password: str = None
    ):
    try:
        _conn = IMAP4_SSL(host)
        _conn.login(account_email, app_password)
        yield _conn
    finally:
        _conn.close()

class EmailEnum(str, Enum):
    ALL = 'ALL'
    INBOX = 'INBOX'
    COPY = 'COPY'
    STORE = 'STORE'
    BODY = '(RFC822)'
    UID = "(UID)"

class EmailAcct:
    def __init__(
        self, 
        account_email: str = None,
        app_password: str = None, 
        which_mailbox: str = 'INBOX',
        host: str = 'imap.gmail.com'
        ):
        
        self.mailbox = which_mailbox

        with IMAP4_conn_manager(
            host=host, 
            account_email=account_email, 
            app_password=app_password) as conn:
            conn.select()
            self.conn = conn

    @property
    def logout(self):
        try:
            self.conn.logout()
        except IMAP4_SSL.error as err:
            raise Exception(f'Logout Error: {err}')

    @classmethod
    def from_gmail(cls, app_password, default_mailbox):
        return cls('imap.gmail.com', app_password, default_mailbox)

    @property
    def list_inboxes(self) -> List[str]:
        box_bytes = self.conn.list()
        decoded_boxes = [item.decode('utf-8') for item in box_bytes[1]]
        return [box.split('" "')[1].replace('"', '') for box in decoded_boxes]

    @property
    def get_namespace(self):
        return self.conn.namespace()

    @property
    def total_emails(self) -> int:
        """Total emails in all mail boxes.

		Returns:
			int: Count of total emails in all inboxes.
		"""
        status, messages = self.conn.select('inbox')
        return int(messages[0])

    def get_recent(self):
        # Prompt server for an update. 
        # Returned data is None if no new messages, 
        # else value of RECENT response.
        new_emails = self.conn.recent()
        return new_emails

    def create_new_mailbox(self, new_mailbox_name):
        '''
        Create new mailbox by name.
        If mailbox exists will fail.

        on failure: ('NO', [b'[ALREADYEXISTS] Duplicate folder name text_mailbox (Failure)'])
        '''
        m = self.conn.create(new_mailbox_name)
        return m

    def delete_emails(self, email_address: str = None):
        # select the mailbox I want to delete in
        # if you want SPAM, use imap.select("SPAM") instead
        self.conn.select("INBOX")
        # search for specific mails by sender
        status, messages = self.conn.search(None, f'(FROM "{email_address}")')
        # # to get mails by subject
        # status, messages = imap.search(None, 'SUBJECT "Thanks for Subscribing to our Newsletter !"')
        # # to get mails after a specific date
        # status, messages = self.conn.search(None, 'SINCE "01-JAN-2020"')
        # # to get mails before a specific date
        # status, messages = imap.search(None, 'BEFORE "01-JAN-2020"')
        # convert messages to a list of email IDs
        messages = messages[0].split(b' ')
        print(messages)
        for mail in messages:
            _, raw_msg = self.conn.fetch(mail, "(RFC822)")

            # you can delete the for loop for performance if you have a long list of emails
            # because it is only for printing the SUBJECT of target email to delete
            for response in raw_msg:
                if isinstance(response, tuple):
                    msg = email.message_from_bytes(response[1])
                    print(msg)
                    # decode the email subject
                    subject = decode_header(msg["Subject"])[0][0]
                    if isinstance(subject, bytes):
                        # if it's a bytes type, decode to str
                        subject = subject.decode()
                print(subject)

        #             print("Deleting", subject)
            # mark the mail as deleted
            # self.conn.store(mail, "+FLAGS", "\\Deleted")
        #     # permanently remove mails that are marked as deleted
        #     # from the selected mailbox (in this case, INBOX)
        #     self.conn.expunge()
        #     # close the mailbox
        #     self.conn.close()
        #     # logout from the account
        #     self.conn.logout()


    def get_unique_sender_emails(
        self, 
        neg_list: Optional[List[Any]] = None,
        which_mailbox: str = 'INBOX'
        ):
        self.conn.list()
        # list of "folders" aka labels in gmail.
        # self.mail.select("[Gmail]/Trash") # connect to inbox.
        self.conn.select(which_mailbox) # connect to inbox.
        # data: [b'1, 2, 3, ..., 1282'] index of all emails in all folders
        result, data = self.conn.search(None, "ALL")
        ids = data[0] # data is a list.
        id_list = ids.split() # ids is a space separated string
        # fetch the email body (RFC822) for the given ID
        unique_senders = set()
        for i in id_list:
            result, data = self.conn.fetch(i, "(RFC822)")
            email_message = email.message_from_bytes(data[0][1])
            unique_senders.add(email_message['From'])
            pprint(unique_senders)
        return unique_senders

    def _search_params(self, params):
        try:
            typ, msg_ids = self.conn.search(None, params)
            if typ == 'OK':
                return msg_ids[0]
            raise Exception('Search Error')
        except Exception as err:
            return err

    def get_raw_email(self, ids):
        id_list = ids.split()# ids is a space separated string
        latest_email_id = id_list[-1] # get the latest
        # fetch the email body (RFC822)
        result, data = self.conn.fetch(latest_email_id, "(RFC822)")
        raw_email = data[0][1].decode("utf-8")
        return raw_email

    def search_emails(self, search_term: str) -> Any:
        self.conn.select()
        email_ids = self._search_params(search_term)
        print(f'{email_ids}')
        # logger.info(f'found: {len(email_ids)} emails')
        # raw_email = self.get_raw_email(email_ids)
        # return raw_email


    def save_as_local(self, filename, payload):
        save_path = os.path.join(f'{os.getcwd()}', f'{datetime.now()}_{filename}')
        fp = open(save_path, 'wb')
        fp.write(payload.get_payload(decode=True))
        fp.close()
        # logger.info(f'filename: {filename}')
        return 'Success'

    def return_attachment_bytes(self, raw_email):
        email_message = email.message_from_string(raw_email)
        for item in email_message.walk():
            filename = item.get_filename()
            # logger.info(f'filename: {filename}')
            if filename:
                file_bytes = item.get_payload(decode=True)
                return file_bytes

    def save_attachment_local(self, raw_email):
        email_message = email.message_from_string(raw_email)
        for item in email_message.walk():
            filename = item.get_filename()
            # logger.info(f'filename: {filename}')
            if filename:
                self.save_as_local(filename, item)


    def get_all_attachments(self, raw_email, as_bytes=True):
        email_message = email.message_from_string(raw_email)
        try:
            for item in email_message.walk():
                filename = item.get_filename()
                # logger.info(f'filename: {filename}')
                if filename:
                    if as_bytes:
                        file_bytes = item.get_payload(decode=True)
                        return file_bytes
                    else:
                        return self.save_as_local(filename, item)
        except Exception as err:
            # logger.info(f'err: {err}')
            return err


    def parse_uid(self, data):
        pattern_uid = re.compile(r'\d+ \(UID (?P<uid>\d+)\)')
        match = pattern_uid.match(data)
        return match.group('uid')
    
    def move_by_email_addr(self, email_address: str, to_mailbox: str = 'text_mailbox'):
        self.conn.select(mailbox='INBOX', readonly=False)
        status, data = self.conn.search(None, f'(FROM "{email_address}")')
        if status == 'OK':
            email_id_list  = data[0].split()
            print(f'{len(email_id_list)} Emails Found for: {email_address}')

            for email_id in email_id_list:
                result, message_id = self.conn.fetch(email_id, "(UID)")
                decoded_id = message_id[0]
                # print(decoded_id)
                if decoded_id:
                    # print(decoded_id.decode('utf-8'))
                    msg_uid = self.parse_uid(decoded_id.decode('utf-8'))
                    result = self.conn.uid('COPY', msg_uid, to_mailbox)
                    if result[0] == 'OK':
                        print(f'Message Copy Success for ID: {msg_uid}')
                        mov, data = self.conn.uid('STORE', msg_uid , '+FLAGS', '(\Deleted)')
                        self.conn.expunge()
                    else:
                        print('Copy Failed')
        else:
            print('Search Failed')
    
    def move_email(self, to_mailbox: str='text_mailbox'):
        self.conn.select(mailbox='inbox', readonly=False)
        resp, items = self.conn.search(None, 'All')
        email_ids  = items[0].split()
        latest_email_id = email_ids[-1]

        # fetch the email body (RFC822) for the given ID
        # result, message = self.conn.fetch(latest_email_id, "(RFC822)")
        # email_message = email.message_from_bytes(message[0][1])
        # print(email_message['Subject'])

        # fetch uid of the latest email
        resp, message_id = self.conn.fetch(latest_email_id, "(UID)")
        decoded_id = message_id[0].decode('utf-8')
        msg_uid = self.parse_uid(decoded_id)
        # print(msg_uid)
        result = self.conn.uid('COPY', msg_uid, to_mailbox)
        if result[0] == 'OK':
            mov, data = self.conn.uid('STORE', msg_uid , '+FLAGS', '(\Deleted)')
            self.conn.expunge()
        else:
            print('Copy Failed')



class Gmail:

    def __init__(self):
        self.mail = IMAP4_SSL('imap.gmail.com')
        self.mail.login('email@gmail.com', 'gmail password')
        self.save_dir = '/path/to/dir'

    def save_as_local(self, filename, payload):
        att_path = os.path.join(self.save_dir, f'{datetime.now()}_{filename}')
        fp = open(att_path, 'wb')
        fp.write(payload.get_payload(decode=True))
        fp.close()
        print('Downloaded file:', filename)

    def save_as_bytes(self, email_message):
        for item in email_message.walk():
            filename = item.get_filename()
            # logger.info(f'filename: {filename}')
            if filename:
                file_bytes = item.get_payload(decode=True)
                return file_bytes


    def save_attachment(self, email_message):
        for item in email_message.walk():
            filename = item.get_filename()
            # logger.info(f'filename: {filename}')
            if filename:
                self.save_as_local(filename, item)


    def search_mail(self):
        self.mail.list()
        # logger.info(f'list: {self.mail.list()}')
        
        # list of "folders" aka labels in gmail.
        # self.mail.select("[Gmail]/Trash") # connect to inbox.
        self.mail.select("inbox") # connect to inbox.
        # result: 'OK'
        # data: [b'1, 2, 3, ..., 1282'] index of all emails in all folders
        # result, data = self.mail.search(None, "ALL")
        # logger.info(data)


    def list_mail(self):
        self.mail.list()
        # logger.info(f'list: {self.mail.list()}')
        
        # list of "folders" aka labels in gmail.
        # self.mail.select("[Gmail]/Trash") # connect to inbox.
        self.mail.select("inbox") # connect to inbox.
        
        # result: 'OK'
        # data: [b'1, 2, 3, ..., 1282'] index of all emails in all folders
        result, data = self.mail.search(None, "ALL")

        ids = data[0] # data is a list.
        # logger.info(f'ids: {data}')

        id_list = ids.split() # ids is a space separated string
        # logger.info(f'id_list: {id_list}')
        
        latest_email_id = id_list[-1] # get the latest
        # logger.info(f'latest_email_id: {latest_email_id}')

        # fetch the email body (RFC822) for the given ID
        result, data = self.mail.fetch(latest_email_id, "(RFC822)") 
        # logger.info(f'data:      {data}')

        raw_email = data[0][1].decode("utf-8")

        email_message = email.message_from_string(raw_email)
        # logger.info(f'email_message: \n\n{email_message}')

        # logger.info(f'email payload: {email_message.keys()}')
        # self.save_attachment(email_message)
        # bts = self.save_as_bytes(email_message)
        # return bts
        return email_message

    def get_attachment_as_bytes(self, email_message):
        # logger.info(f'email payload: {email_message.keys()}')
        bts = self.save_as_bytes(email_message)
        return bts
